from typing import List, Callable, Optional, Union, Dict
from copy import deepcopy
import numpy as np

from .qaoa_result import QAOAResult
from ..workflow_properties import CircuitProperties
from ..baseworkflow import Workflow, check_compiled
from ...backends import QAOABackendAnalyticalSimulator
from ...backends.devices_core import DeviceLocal
from ...backends.qaoa_backend import get_qaoa_backend
from ...problems import QUBO
from ...qaoa_components import (
    Hamiltonian,
    QAOADescriptor,
    create_qaoa_variational_params,
)
from ...qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from ...utilities import (
    get_mixer_hamiltonian,
    generate_timestamp,
    ground_state_hamiltonian,
)
from ...optimizers.qaoa_optimizer import get_optimizer
from ...backends.wrapper import SPAMTwirlingWrapper


class QAOA(Workflow):
    """
    A class implementing a QAOA workflow end to end.

    It's basic usage consists of
    1. Initialization
    2. Compilation
    3. Optimization

    .. note::
        The attributes of the QAOA class should be initialized using the set methods of QAOA.
        For example, to set the circuit's depth to 10 you should run `set_circuit_properties(p=10)`

    Attributes
    ----------
    device: `DeviceBase`
        Device to be used by the optimizer
    circuit_properties: `CircuitProperties`
        The circuit properties of the QAOA workflow. Use to set depth `p`,
        choice of parameterization, parameter initialisation strategies, mixer hamiltonians.
        For a complete list of its parameters and usage please see the method `set_circuit_properties`
    backend_properties: `BackendProperties`
        The backend properties of the QAOA workflow. Use to set the backend properties
        such as the number of shots and the cvar values.
        For a complete list of its parameters and usage please see the method `set_backend_properties`
    classical_optimizer: `ClassicalOptimizer`
        The classical optimiser properties of the QAOA workflow. Use to set the
        classical optimiser needed for the classical optimisation part of the QAOA routine.
        For a complete list of its parameters and usage please see the method `set_classical_optimizer`
    local_simulators: `list[str]`
        A list containing the available local simulators
    cloud_provider: `list[str]`
        A list containing the available cloud providers
    mixer_hamil: Hamiltonian
        The desired mixer hamiltonian
    cost_hamil: Hamiltonian
        The desired mixer hamiltonian
    qaoa_descriptor: QAOADescriptor
        the abstract and backend-agnostic representation of the underlying QAOA parameters
    variate_params: QAOAVariationalBaseParams
        The variational parameters. These are the parameters to be optimised by the classical optimiser
    backend: VQABaseBackend
        The openQAOA representation of the backend to be used to execute the quantum circuit
    optimizer: OptimizeVQA
        The classical optimiser
    result: `Result`
        Contains the logs of the optimisation process
    compiled: `Bool`
        A boolean flag to check whether the QAOA object has been correctly compiled at least once

    Examples
    --------
    Examples should be written in doctest format, and should illustrate how
    to use the function.

    >>> q = QAOA()
    >>> q.compile(QUBO)
    >>> q.optimize()

    Where `QUBO` is a an instance of `openqaoa.problems.problem.QUBO`

    If you want to use non-default parameters:

    >>> q_custom = QAOA()
    >>> q_custom.set_circuit_properties(
            p=10,
            param_type='extended',
            init_type='ramp',
            mixer_hamiltonian='x'
        )
    >>> q_custom.set_device_properties(
            device_location = 'qcs', device_name='Aspen-11',
            cloud_credentials = {
                'name' : "Aspen11", 'as_qvm':True,
                'execution_timeout' : 10, 'compiler_timeout':10
            }
        )
    >>> q_custom.set_backend_properties(n_shots=200, cvar_alpha=1)
    >>> q_custom.set_classical_optimizer(method='nelder-mead', maxiter=2)
    >>> q_custom.compile(qubo_problem)
    >>> q_custom.optimize()
    """

    results_class = QAOAResult

    def __init__(self, device=DeviceLocal("vectorized")):
        """
        Initialize the QAOA class.

        Parameters
        ----------
        device: `DeviceBase`
            Device to be used by the optimizer. Default is using the local 'vectorized' simulator.
        """
        super().__init__(device)
        self.circuit_properties = CircuitProperties()

        # change header algorithm to qaoa
        self.header["algorithm"] = "qaoa"

    @check_compiled
    def set_circuit_properties(self, **kwargs):
        """
        Specify the circuit properties to construct QAOA circuit

        Parameters
        ----------
        qubit_register: `list`
            Select the desired qubits to run the QAOA program. Meant to be used as a qubit
            selector for qubits on a QPU. Defaults to a list from 0 to n-1 (n = number of qubits)
        p: `int`
            Depth `p` of the QAOA circuit
        q: `int`
            Analogue of `p` of the QAOA circuit in the Fourier parameterization
        param_type: `str`
            Choose the QAOA circuit parameterization. Currently supported parameterizations include:
            `'standard'`: Standard QAOA parameterization
            `'standard_w_bias'`: Standard QAOA parameterization with a separate parameter for single-qubit terms.
            `'extended'`: Individual parameter for each qubit and each term in the Hamiltonian.
            `'fourier'`: Fourier circuit parameterization
            `'fourier_extended'`: Fourier circuit parameterization with individual parameter
            for each qubit and term in Hamiltonian.
            `'fourier_w_bias'`: Fourier circuit parameterization with a separate
            parameter for single-qubit terms
        init_type: `str`
            Initialisation strategy for the QAOA circuit parameters. Allowed init_types:
            `'rand'`: Randomly initialise circuit parameters
            `'ramp'`: Linear ramp from Hamiltonian initialisation of circuit
            parameters (inspired from Quantum Annealing)
            `'custom'`: User specified initial circuit parameters
        mixer_hamiltonian: `str`
            Parameterisation of the mixer hamiltonian:
            `'x'`: Randomly initialise circuit parameters
            `'xy'`: Linear ramp from Hamiltonian initialisation of circuit
        mixer_qubit_connectivity: `[Union[List[list],List[tuple], str]]`
            The connectivity of the qubits in the mixer Hamiltonian. Use only if
            `mixer_hamiltonian = xy`. The user can specify the connectivity as a list of lists,
            a list of tuples, or a string chosen from ['full', 'chain', 'star'].
        mixer_coeffs: `list`
            The coefficients of the mixer Hamiltonian. By default all set to -1
        annealing_time: `float`
            Total time to run the QAOA program in the Annealing parameterization (digitised annealing)
        linear_ramp_time: `float`
            The slope(rate) of linear ramp initialisation of QAOA parameters.
        variational_params_dict: `dict`
            Dictionary object specifying the initial value of each circuit parameter for
            the chosen parameterization, if the `init_type` is selected as `'custom'`.
            For example, for standard params set {'betas': [0.1, 0.2, 0.3], 'gammas': [0.1, 0.2, 0.3]}
        """

        for key, value in kwargs.items():
            if hasattr(self.circuit_properties, key):
                pass
            else:
                raise ValueError("Specified argument is not supported by the circuit")
        self.circuit_properties = CircuitProperties(**kwargs)

        return None

    def compile(
        self,
        problem: QUBO = None,
        verbose: bool = False,
        routing_function: Optional[Callable] = None,
    ):
        """
        Initialise the trainable parameters for QAOA according to the specified
        strategies and by passing the problem statement

        .. note::
            Compilation is necessary because it is the moment where the problem statement and
            the QAOA instructions are used to build the actual QAOA circuit.

        .. tip::
            Set Verbose to false if you are running batch computations!

        Parameters
        ----------
        problem: `Problem`
            QUBO problem to be solved by QAOA
        verbose: bool
            Set True to have a summary of QAOA to displayed after compilation
        """
        # if isinstance(routing_function,Callable):
        #     #assert that routing_function is supported only for Standard QAOA.
        #     if (
        #         self.backend_properties.append_state is not None or\
        #         self.backend_properties.prepend_state is not None or\
        #         self.circuit_properties.mixer_hamiltonian is not 'x' or\

        #     )

        # connect to the QPU specified
        self.device.check_connection()
        # we compile the method of the parent class to genereate the id and
        # check the problem is a QUBO object and save it
        super().compile(problem=problem)

        self.cost_hamil = Hamiltonian.classical_hamiltonian(
            terms=problem.terms, coeffs=problem.weights, constant=problem.constant
        )

        self.mixer_hamil = get_mixer_hamiltonian(
            n_qubits=self.cost_hamil.n_qubits,
            mixer_type=self.circuit_properties.mixer_hamiltonian,
            qubit_connectivity=self.circuit_properties.mixer_qubit_connectivity,
            coeffs=self.circuit_properties.mixer_coeffs,
        )

        self.qaoa_descriptor = QAOADescriptor(
            self.cost_hamil,
            self.mixer_hamil,
            p=self.circuit_properties.p,
            routing_function=routing_function,
            device=self.device,
        )
        self.variate_params = create_qaoa_variational_params(
            qaoa_descriptor=self.qaoa_descriptor,
            params_type=self.circuit_properties.param_type,
            init_type=self.circuit_properties.init_type,
            variational_params_dict=self.circuit_properties.variational_params_dict,
            linear_ramp_time=self.circuit_properties.linear_ramp_time,
            q=self.circuit_properties.q,
            seed=self.circuit_properties.seed,
            total_annealing_time=self.circuit_properties.annealing_time,
        )

        backend_dict = self.backend_properties.__dict__.copy()

        self.backend = get_qaoa_backend(
            qaoa_descriptor=self.qaoa_descriptor,
            device=self.device,
            **backend_dict,
        )

        # Implementing SPAM Twirling error mitigation requires wrapping the backend.
        # However, the BaseWrapper can have many more use cases.
        if (
            self.error_mitigation_properties.error_mitigation_technique
            == "spam_twirling"
        ):
            self.backend = SPAMTwirlingWrapper(
                backend=self.backend,
                n_batches=self.error_mitigation_properties.n_batches,
                calibration_data_location=self.error_mitigation_properties.calibration_data_location,
            )

        self.optimizer = get_optimizer(
            vqa_object=self.backend,
            variational_params=self.variate_params,
            optimizer_dict=self.classical_optimizer.asdict(),
        )

        # Set the header properties
        self.header["target"] = self.device.device_name
        self.header["cloud"] = self.device.device_location

        metadata = {
            "p": self.circuit_properties.p,
            "param_type": self.circuit_properties.param_type,
            "init_type": self.circuit_properties.init_type,
            "optimizer_method": self.classical_optimizer.method,
        }

        self.set_exp_tags(tags=metadata)

        self.compiled = True

        if verbose:
            print("\t \033[1m ### Summary ###\033[0m")
            print("OpenQAOA has been compiled with the following properties")
            print(
                f"Solving QAOA with \033[1m {self.device.device_name} \033[0m on"
                f"\033[1m{self.device.device_location}\033[0m"
            )
            print(
                f"Using p={self.circuit_properties.p} with {self.circuit_properties.param_type}"
                f"parameters initialized as {self.circuit_properties.init_type}"
            )

            if hasattr(self.backend, "n_shots"):
                print(
                    f"OpenQAOA will optimize using \033[1m{self.classical_optimizer.method}"
                    f"\033[0m, with up to \033[1m{self.classical_optimizer.maxiter}"
                    f"\033[0m maximum iterations. Each iteration will contain"
                    f"\033[1m{self.backend_properties.n_shots} shots\033[0m"
                )
            else:
                print(
                    f"OpenQAOA will optimize using \033[1m{self.classical_optimizer.method}\033[0m,"
                    "with up to \033[1m{self.classical_optimizer.maxiter}\033[0m maximum iterations"
                )

        return None

    def solve_brute_force(self, bounded=True, verbose=False):
        """
        A method to solve the QAOA problem using brute force i.e. by
        evaluating the cost function at all possible bitstrings

        Parameters
        ----------
        bounded: `bool`, optional
            If set to True, the function will not perform computations for qubit
            numbers above 25. If False, the user can specify any number. Defaults
            to True.
        verbose: `bool`, optional
            If set to True, the function will print the results of the computation.
            Defaults to False.
        """

        if self.compiled is False:
            raise ValueError(
                "Please compile the QAOA before running the brute force solver!"
            )

        # compute the exact ground state and ground state energy of the cost hamiltonian
        energy, configuration = ground_state_hamiltonian(
            self.cost_hamil, bounded=bounded
        )

        if verbose:
            print(f"Ground State energy: {energy}, Solution: {configuration}")

        self.brute_force_results = {
            "energy": energy,
            "configuration": configuration,
        }

        return None

    def optimize(self, verbose=False):
        """
        A method running the classical optimisation loop
        """

        if self.compiled is False:
            raise ValueError("Please compile the QAOA before optimizing it!")

        # timestamp for the start of the optimization
        self.header["execution_time_start"] = generate_timestamp()

        self.optimizer.optimize()
        # TODO: result and qaoa_result will differ
        self.result = self.optimizer.qaoa_result

        # timestamp for the end of the optimization
        self.header["execution_time_end"] = generate_timestamp()

        if verbose:
            print("Optimization completed.")
        return

    def evaluate_circuit(
        self,
        params: Union[List[float], Dict[str, List[float]], QAOAVariationalBaseParams],
    ):
        """
        A method to evaluate the QAOA circuit at a given set of parameters

        Parameters
        ----------
        params: list or dict or QAOAVariationalBaseParams or None
            List of parameters or dictionary of parameters. Which will be used to evaluate the QAOA circuit.
            If None, the variational parameters of the QAOA object will be used.

        Returns
        -------
        result: dict
            A dictionary containing the results of the evaluation:
            - "cost": the expectation value of the cost Hamiltonian
            - "uncertainty": the uncertainty of the expectation value of the cost Hamiltonian
            - "measurement_results": either the state of the QAOA circuit output (if the QAOA circuit is
            evaluated on a state simulator) or the counts of the QAOA circuit output
            (if the QAOA circuit is evaluated on a QPU or shot-based simulator)
        """

        # before evaluating the circuit we check that the QAOA object has been compiled
        if self.compiled is False:
            raise ValueError("Please compile the QAOA before optimizing it!")

        # Check the type of the input parameters and save them as a
        # QAOAVariationalBaseParams object at the variable `params_obj`

        # if the parameters are passed as a dictionary we copy and update the variational parameters of the QAOA object
        if isinstance(params, dict):
            params_obj = deepcopy(self.variate_params)
            # we check that the dictionary contains all the parameters of the QAOA object that are not empty
            for key, value in params_obj.asdict().items():
                if value.size > 0:
                    assert (
                        key in params.keys()
                    ), f"The parameter `{key}` is missing from the input dictionary"
            params_obj.update_from_dict(params)

        # if the parameters are passed as a list we copy and update the variational parameters of the QAOA object
        elif isinstance(params, list) or isinstance(params, np.ndarray):
            assert len(params) == len(
                self.variate_params
            ), "The number of parameters does not match the number of parameters in the QAOA circuit"
            params_obj = deepcopy(self.variate_params)
            params_obj.update_from_raw(params)

        # if the parameters are passed as a QAOAVariationalBaseParams object we just take it as it is
        elif isinstance(params, QAOAVariationalBaseParams):
            # check whether the input params object is supported for circuit evaluation
            assert (
                len(self.variate_params.mixer_1q_angles) == len(params.mixer_1q_angles)
                and len(self.variate_params.mixer_2q_angles)
                == len(self.variate_params.mixer_2q_angles)
                and len(self.variate_params.cost_1q_angles)
                == len(self.variate_params.cost_1q_angles)
                and len(self.variate_params.cost_2q_angles)
                == len(self.variate_params.cost_2q_angles)
            ), "Specify a supported params object"
            params_obj = params

        # if the parameters are passed in a different format, we raise an error
        else:
            raise TypeError(
                f"The input params must be a list or a dictionary. Instead, received {type(params)}"
            )

        # Evaluate the QAOA circuit and return the results
        output_dict = {
            "cost": None,
            "uncertainty": None,
            "measurement_results": None,
        }
        # if the backend is the analytical simulator, we just return the expectation value of the cost Hamiltonian
        if isinstance(self.backend, QAOABackendAnalyticalSimulator):
            output_dict.update({"cost": self.backend.expectation(params_obj)[0]})

        # if the workflow implements SPAM Twirling,
        # we just return the expectation value of the cost Hamiltonian and the measurement outcomes
        elif isinstance(self.backend, SPAMTwirlingWrapper):
            cost = self.backend.expectation(params_obj)
            measurement_results = (
                self.backend.measurement_outcomes
                if isinstance(self.backend.measurement_outcomes, dict)
                else self.backend.measurement_outcomes.tolist()
            )
            output_dict.update(
                {
                    "cost": cost,
                    "measurement_results": measurement_results,
                }
            )
            # in all other cases, we return the expectation value of the cost Hamiltonian,
            # the associated uncertainty and the measurement outcomes
        else:
            cost, uncertainty = self.backend.expectation_w_uncertainty(params_obj)
            measurement_results = (
                self.backend.measurement_outcomes
                if isinstance(self.backend.measurement_outcomes, dict)
                else self.backend.measurement_outcomes.tolist()
            )
            output_dict.update(
                {
                    "cost": cost,
                    "uncertainty": uncertainty,
                    "measurement_results": measurement_results,
                }
            )
        return output_dict

    def _serializable_dict(
        self, complex_to_string: bool = False, intermediate_measurements: bool = True
    ):
        """
        Returns all values and attributes of the object that we want to return in
        `asdict` and `dump(s)` methods in a dictionary.

        Parameters
        ----------
        complex_to_string: bool
            If True, complex numbers are converted to strings.
            This is useful for JSON serialization.

        Returns
        -------
        serializable_dict: dict
            A dictionary containing all the values and attributes of the object
            that we want to return in `asdict` and `dump(s)` methods.
        intermediate_measurements: bool
            If True, intermediate measurements are included in the dump.
            If False, intermediate measurements are not included in the dump.
            Default is True.
        """

        # we call the _serializable_dict method of the parent class,
        # specifying the keys to delete from the results dictionary
        serializable_dict = super()._serializable_dict(
            complex_to_string, intermediate_measurements
        )

        # we add the keys of the QAOA object that we want to return
        serializable_dict["data"]["input_parameters"]["circuit_properties"] = dict(
            self.circuit_properties
        )

        # include parameters in the header metadata
        serializable_dict["header"]["metadata"]["param_type"] = serializable_dict[
            "data"
        ]["input_parameters"]["circuit_properties"]["param_type"]
        serializable_dict["header"]["metadata"]["init_type"] = serializable_dict[
            "data"
        ]["input_parameters"]["circuit_properties"]["init_type"]
        serializable_dict["header"]["metadata"]["p"] = serializable_dict["data"][
            "input_parameters"
        ]["circuit_properties"]["p"]

        if (
            serializable_dict["data"]["input_parameters"]["circuit_properties"]["q"]
            is not None
        ):
            serializable_dict["header"]["metadata"]["q"] = serializable_dict["data"][
                "input_parameters"
            ]["circuit_properties"]["q"]

        return serializable_dict



import numpy as np
from sympy import Symbol
import unittest

from openqaoa.qaoa_components import PauliOp, Hamiltonian

"""
Unittest based testing of the PauliOp and Hamiltonian classes
"""


class TestingOperators(unittest.TestCase):
    def test_pauli_init(self):
        """
        Test the initialization method of the PauliOp class.

        The test consists in checking correct initializations for representative examples.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 1),
            (4, 5, 0, 1),
            (2, 0, 7, 4, 5),
            (0, 1, 5, 3, 7, 6),
            (6, 5, 4, 3, 1, 1, 0),
        ]
        input_strings = ["XZX", "YYYY", "ZYXZX", "YYZZXX", "ZYXYZYX"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Correctly ordered indices and strings
        correct_indices = [
            (0, 1),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 3, 4, 5, 6),
        ]
        correct_strings = ["XY", "YYYY", "YZZXX", "YYZZXX", "XXYXYZ"]
        correct_phases = [1j, 1, 1, 1, -1j]

        # Extract indices and strings generated by the Pauli operators
        indices = [pauli.qubit_indices for pauli in paulis]
        strings = [pauli.pauli_str for pauli in paulis]
        phases = [pauli.phase for pauli in paulis]

        # Test that indices and strings were generated correctly
        assert indices == correct_indices, f"Pauli indices not generated correctly"
        assert strings == correct_strings, f"Pauli strings not generated correctly"
        assert np.allclose(
            phases, correct_phases
        ), f"Pauli phases not generated correctly"

        # Exception cases - not valid Pauli operators
        input_exc_indices = [
            (0, 2, 1),
            (1,),
            (2, 2, 2, 4, 5),
            (0, 1, 5, 3, 7, 6),
            (6, 5, 4, 3, 2, 1, 0),
        ]
        input_exc_strings = ["XY", "XZZYZXYZ", "YXZZ", "XYZZX", "Z"]

        # Test the exceptions are raised
        for string, index in zip(input_exc_strings, input_exc_indices):
            # Attempt construction of Pauli operator
            with self.assertRaises(AssertionError) as context:
                pauli = PauliOp(string, index)

            # Check exception message
            self.assertEqual(
                f"Each Pauli operator must have a unique qubit index",
                str(context.exception),
            )

    def test_sort_pauli_op(self):
        """
        Test of the function that sorts the string attribute and the indices
        of the pauli operator object.

        The test consists in properly ordering a set of string + indices.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 2, 1),
            (4, 5, 0, 1),
            (2, 0, 7, 4, 5),
            (0, 1, 5, 3, 7, 6),
            (6, 5, 4, 3, 2, 1, 0),
        ]
        input_strings = ["XXZ", "YYYY", "ZYXZX", "YYZZXX", "ZYXYZYX"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Correctly ordered indices and strings
        correct_indices = [
            (0, 1, 2),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 2, 3, 4, 5, 6),
        ]
        correct_strings = ["XZX", "YYYY", "YZZXX", "YYZZXX", "XYZYXYZ"]

        # Extract indices and strings generated by the Pauli operators
        indices = [pauli.qubit_indices for pauli in paulis]
        strings = [pauli.pauli_str for pauli in paulis]

        # Test that indices and strings were correctly ordered when creating the Pauli Operator
        assert indices == correct_indices, f"Pauli indices not sorted correctly"
        assert strings == correct_strings, f"Pauli strings not sorted correctly"

        # Exception cases - not valid Pauli operators
        input_exc_indices = [
            (0, 2, 1),
            (1,),
            (2, 0, 7, 4, 5),
            (0, 1, 5, 3, 7, 6),
            (6, 5, 4, 3, 2, 1, 0),
        ]
        input_exc_strings = ["LXY", "Q", "ZYXZR", "FYZZXX", "ZYXYZYN"]

        # Pauli set
        PAULIS_SET = set("XYZI")

        # Test the exceptions are raised
        for string, index in zip(input_exc_strings, input_exc_indices):
            # Extract exception from string
            non_pauli_operator = list(set(string) - PAULIS_SET)[0]

            # Attempt construction of Pauli operator
            with self.assertRaises(ValueError) as context:
                pauli = PauliOp(string, index)

            # Check exception message
            self.assertEqual(
                f"{non_pauli_operator} is not a valid Pauli. Please choose from the set {PAULIS_SET}",
                str(context.exception),
            )

    def test_pauli_simplify(self):
        """
        Test of the function that simplifies the expression of Pauli Operator by using
        the multiplication rules that Pauli matrices obey.

        The test consists in simplify a set of Pauli operators.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 1),
            (0, 0, 0, 2),
            (0, 1, 3, 3, 4),
            (1, 1, 1, 1, 1, 1, 1),
        ]
        input_strings = ["ZXY", "YYYY", "ZYIZX", "ZZZZZXZ"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Indices, strings and phases from correctly simplified Pauli operators
        correct_indices = [(0, 1), (0, 2), (0, 1, 3, 4), (1,)]
        correct_strings = ["ZZ", "YY", "ZYZX", "X"]
        correct_phases = [1j, 1, 1, -1]

        # Extract indices, strings and phases generated by the Pauli operators
        indices = [pauli.qubit_indices for pauli in paulis]
        strings = [pauli.pauli_str for pauli in paulis]
        phases = [pauli.phase for pauli in paulis]

        # Test that indices and strings were correctly ordered when creating the Pauli Operator
        assert indices == correct_indices, f"Indices not simplified correctly"
        assert strings == correct_strings, f"Strings not simplified correctly"
        assert np.allclose(phases, correct_phases), f"Phases not simplified correctly"

    def test_pauli_is_trivial(self):
        """
        Test of the function that checks if a Pauli Operator object is trivial,
        i.e. only contains identity operators.

        The tests consists in checking the triviality of a set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 2),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 2, 3, 4, 5, 6),
        ]
        input_strings = ["III", "IYYY", "YZIXX", "IIIIII", "XIIIIII"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Correct list containing whether the Pauli Operator is trivial or not
        correct_triviality = [True, False, False, True, False]

        # Compute triviality of the Pauli operators
        triviality = [pauli._is_trivial for pauli in paulis]

        # Test that indices and strings were correctly ordered when creating the Pauli Operator
        assert np.allclose(
            triviality, correct_triviality
        ), f"Triviality classification is incorrect"

    def test_pauli_matrix(self):
        """
        Test of the function that computes the matrix representation of the Pauli operator.

        The test consists in computing the matrix representation of set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1), (0, 1), (0, 1), (0, 1), (0, 1, 2)]
        input_strings = ["IX", "XY", "XZ", "ZY", "IIZ"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Correct pauli representations
        correct_matrix_representations = [
            np.array(
                ([0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]), dtype=complex
            ),
            np.array(
                ([0, 0, 0, -1j], [0, 0, 1j, 0], [0, -1j, 0, 0], [1j, 0, 0, 0]),
                dtype=complex,
            ),
            np.array(
                ([0, 0, 1, 0], [0, 0, 0, -1], [1, 0, 0, 0], [0, -1, 0, 0]),
                dtype=complex,
            ),
            np.array(
                ([0, -1j, 0, 0], [1j, 0, 0, 0], [0, 0, 0, 1j], [0, 0, -1j, 0]),
                dtype=complex,
            ),
            np.array(
                (
                    [1, 0, 0, 0, 0, 0, 0, 0],
                    [0, -1, 0, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 0, -1, 0, 0, 0, 0],
                    [0, 0, 0, 0, 1, 0, 0, 0],
                    [0, 0, 0, 0, 0, -1, 0, 0],
                    [0, 0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 0, -1],
                ),
                dtype=complex,
            ),
        ]

        # Compute matrix representations
        matrix_representations = [pauli.matrix for pauli in paulis]

        # Compare each matrix representation
        matrix_comparison = [
            (correct_matrix_representations[i] == matrix_representations[i]).all()
            for i in range(len(matrix_representations))
        ]

        # Test that the computed representations are correct
        assert np.all(matrix_comparison), f"Matrix representation is incorrect"

    def test_pauli_len(self):
        """
        Test the function that computes the length of the Pauli operator.

        The tests consists in computing the length for several examples.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 2),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 2, 3, 4, 5, 6),
        ]
        input_strings = ["III", "IYYY", "YZIXX", "IIIIII", "XIIIIII"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Correct Pauli operator lengths
        correct_lengths = [3, 4, 5, 6, 7]

        # Compute Pauli operator lengths
        lengths = [pauli.__len__() for pauli in paulis]

        # Test that computed lengths are correct
        assert np.allclose(lengths, correct_lengths), f"Computed lengths are incorrect"

    def test_pauli_eq(self):
        """
        Test the function that checks whether two Pauli operators are equivalent.

        The test consists in checking equivalence between different Pauli operators for several examples.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 2),
            (0, 1, 2),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 2, 3, 4, 5, 6),
        ]
        input_strings = ["XZX", "XZX", "YYYY", "YZZXX", "YYZZXX", "YYZZXX", "XYZYXYZ"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Check equivalence between contiguous Pauli operators in the list
        equivalences = [paulis[i].__eq__(paulis[i + 1]) for i in range(len(paulis) - 1)]

        # Correct equivalences
        correct_equivalences = [True, False, False, False, True, False]

        # Test the checks have been performes correctly
        assert np.allclose(
            equivalences, correct_equivalences
        ), f"Equivalences have not been computed correctly"

    def test_pauli_copy(self):
        """
        Test the function that computes a copy of a Pauli operator.

        The test consists in creating copies for a set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 2),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 2, 3, 4, 5, 6),
        ]
        input_strings = ["XZX", "YYYY", "YZZXX", "YYZZXX", "XYZYXYZ"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Generate copies of the Pauli operators
        copy_paulis = [pauli.__copy__() for pauli in paulis]

        # Test the copies have been correctly performes

        check_copy = [paulis[j].__eq__(copy_paulis[j]) for j in range(len(paulis))]

        assert sum(check_copy) == len(
            check_copy
        ), f"Copies were not performed correctly"

    def test_pauli_str(self):
        """
        Test the function that generates a string representation of the Pauli operator.

        The test consists in computing the string representation of a set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 2),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 2, 3, 4, 5, 6),
        ]
        input_strings = ["XZX", "YYYY", "YZZXX", "YYZZXX", "XYZYXYZ"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Correct string representations
        correct_str_rep = [
            "X_{0}Z_{1}X_{2}",
            "Y_{0}Y_{1}Y_{4}Y_{5}",
            "Y_{0}Z_{2}Z_{4}X_{5}X_{7}",
            "Y_{0}Y_{1}Z_{3}Z_{5}X_{6}X_{7}",
            "X_{0}Y_{1}Z_{2}Y_{3}X_{4}Y_{5}Z_{6}",
        ]

        # Generate string rtepresentations
        str_rep = [pauli.__str__() for pauli in paulis]

        # Test that string representations were correctly generated
        assert (
            str_rep == correct_str_rep
        ), f"String representations were not correctly generated"

    def test_pauli_repr(self):
        """
        Test the function that generates a string representation of the Pauli operator object.

        The test consists in computing the string representation of a set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 2),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 2, 3, 4, 5, 6),
        ]
        input_strings = ["XZX", "YYYY", "YZZXX", "YYZZXX", "XYZYXYZ"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Correct string representations
        correct_str_rep = [
            "PauliOp(XZX,(0, 1, 2))",
            "PauliOp(YYYY,(0, 1, 4, 5))",
            "PauliOp(YZZXX,(0, 2, 4, 5, 7))",
            "PauliOp(YYZZXX,(0, 1, 3, 5, 6, 7))",
            "PauliOp(XYZYXYZ,(0, 1, 2, 3, 4, 5, 6))",
        ]

        # Generate string rtepresentations
        str_rep = [pauli.__repr__() for pauli in paulis]

        # Test that string representations were correctly generated
        assert (
            str_rep == correct_str_rep
        ), f"String representations for Pauli operator object were not correctly generated"

    def test_pauli_mul(self):
        """
        Test the functions that multiply two Pauli operators.

        The test consists in multipliying a set of example Pauli operators.
        """

        # Define input examples of indices and strings
        input_indices = [
            (0, 1, 2),
            (0, 1, 4, 5),
            (0, 2, 4, 5, 7),
            (0, 1, 3, 5, 6, 7),
            (0, 1, 2, 3, 4, 5, 6),
        ]
        input_strings = ["XZX", "YYYY", "YZZXX", "YYZZXX", "XYZYXYX"]

        # Create Pauli operators from input examples
        paulis = [
            PauliOp(string, index)
            for string, index in zip(input_strings, input_indices)
        ]

        # Compute multiplication between contiguous Pauli operators in the list
        pauli_mult = [paulis[j].__mul__(paulis[j + 1]) for j in range(len(paulis) - 1)]
        indices = [pauli.qubit_indices for pauli in pauli_mult]
        strings = [pauli.pauli_str for pauli in pauli_mult]
        phases = [pauli.phase for pauli in pauli_mult]

        # Correct matrix multiplications
        correct_indices = [
            (0, 1, 2, 4, 5),
            (1, 2, 4, 5, 7),
            (1, 2, 3, 4, 5, 6),
            (0, 2, 3, 4, 5, 7),
        ]
        correct_strings = ["ZXXYY", "YZXZX", "YZZZYX", "ZZXXXX"]
        correct_phases = [1, 1, -1j, 1j]

        # Test that matrix multiplications are correct
        assert (
            indices == correct_indices
        ), f"Indices from matrix multiplications are not correct"
        assert (
            strings == correct_strings
        ), f"Strings from matrix multiplications are not correct"
        assert np.allclose(
            phases, correct_phases
        ), f"Phases from matrix multiplications are not correct"

    def test_pauli_XYZI(self):
        """
        Test the functions that define each Pauli matrix.

        The test consists in checking the properties of each Pauli matrix when defined as a Pauli
        operator object.
        """

        # Define a qubit index to define the Pauli operator object
        index = 1

        # Define the Pauli matrices
        X = PauliOp.X(index)
        Y = PauliOp.Y(index)
        Z = PauliOp.Z(index)
        I = PauliOp.I(index)
        paulis = [X, Y, Z, I]

        # Extract attributes
        strings = [pauli.pauli_str for pauli in paulis]
        indices = [pauli.qubit_indices for pauli in paulis]
        phases = [pauli.phase for pauli in paulis]
        string_rep = [pauli.__str__() for pauli in paulis]
        string_obj_rep = [pauli.__repr__() for pauli in paulis]
        matrix_rep = [pauli.matrix for pauli in paulis]

        # Correct attributes
        correct_strings = ["X", "Y", "Z", "I"]
        correct_indices = [(1,), (1,), (1,), (1,)]
        correct_phases = [1, 1, 1, 1]
        correct_string_rep = ["X_{1}", "Y_{1}", "Z_{1}", "I_{1}"]
        correct_string_obj_rep = [
            "PauliOp(X,(1,))",
            "PauliOp(Y,(1,))",
            "PauliOp(Z,(1,))",
            "PauliOp(I,(1,))",
        ]
        correct_matrix_rep = [
            np.array(([0, 1], [1, 0]), dtype=complex),
            np.array(([0, -1j], [1j, 0]), dtype=complex),
            np.array(([1, 0], [0, -1]), dtype=complex),
            np.array(([1, 0], [0, 1]), dtype=complex),
        ]

        # Check indices, strings, phases and matrix representations
        assert (
            strings == correct_strings
        ), f"Strings are not correctly defined for Pauli matrices"
        assert (
            indices == correct_indices
        ), f"Indices are not correctly defined for Pauli matrices"
        assert np.allclose(
            phases, correct_phases
        ), f"Phases are not correctly defined for Pauli matrices"
        assert (
            string_rep == correct_string_rep
        ), f"String representations are not correctly defined for Pauli matrices"
        assert (
            string_obj_rep == correct_string_obj_rep
        ), f"String object representations are not correctly defined for Pauli matrices"
        assert np.all(
            [
                (matrix_rep[i] == correct_matrix_rep[i]).all()
                for i in range(len(matrix_rep))
            ]
        ), f"Matrix representations are not correctly defined for Pauli matrices"

    def test_hamiltonian_init(self):
        """
        Test initialization method of the Hamiltonian class.

        The tests consists in checking correct initializations for representative examples.
        """

        # Number of qubits in the register
        n_qubits = 3

        ## First example

        # Define Pauli terms using shifted indices
        idx_ref = 3
        indices_ref = list(range(idx_ref, n_qubits + idx_ref))
        trivial_indices = [(i,) for i in indices_ref]
        singlet_indices = [(i,) for i in indices_ref]
        pair_indices = [(i, j) for j in indices_ref for i in range(idx_ref, j)]

        trivial_terms = [PauliOp("I", indices) for indices in trivial_indices]
        linear_terms = [PauliOp("Z", indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp("ZZ", indices) for indices in pair_indices]

        terms = trivial_terms + linear_terms + quadratic_terms

        # Define coefficients and constant
        trivial_coeffs = [0.5 for _ in range(len(trivial_indices))]
        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        coefficients = trivial_coeffs + linear_coeffs + quadratic_coeffs

        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Correct attributes
        correct_indices = list(range(n_qubits))
        correct_singlet_indices = [(i,) for i in correct_indices]
        correct_pair_indices = [(i, j) for j in correct_indices for i in range(j)]
        correct_linear_terms = [
            PauliOp("Z", indices) for indices in correct_singlet_indices
        ]
        correct_quadratic_terms = [
            PauliOp("ZZ", indices) for indices in correct_pair_indices
        ]
        correct_linear_coeffs = [-1 for _ in range(len(correct_singlet_indices))]
        correct_quadratic_coeffs = [0.5 for _ in range(len(correct_pair_indices))]

        correct_terms = correct_linear_terms + correct_quadratic_terms
        correct_coefficients = correct_linear_coeffs + correct_quadratic_coeffs
        correct_constant = constant + sum(trivial_coeffs)

        # Test Hamiltonian was defined correctly
        assert (
            hamiltonian.terms == correct_terms
        ), f"Terms in the Hamiltonian were not constructed correctly"
        assert (
            hamiltonian.coeffs == correct_coefficients
        ), f"Coefficients in the Hamiltonian were not constructed correctly"
        assert (
            hamiltonian.constant == correct_constant
        ), f"Constant in the Hamiltonian was not correctly constructed"

        ## Second example

        # Define Pauli terms
        trivial_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]

        trivial_terms = [PauliOp("I", indices) for indices in trivial_indices]
        terms_strings = ["XX", "YY"]
        quadratic_terms = [
            PauliOp(string, indices)
            for indices in pair_indices
            for string in terms_strings
        ]

        terms = trivial_terms + quadratic_terms

        # Define coefficients and constant
        trivial_coeffs = [-1 for _ in range(len(trivial_terms))]
        quadratic_coeffs = [0.5 for _ in range(len(quadratic_terms))]
        coefficients = trivial_coeffs + quadratic_coeffs

        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Correct attributes
        correct_terms = quadratic_terms
        correct_coefficients = quadratic_coeffs
        correct_constant = constant + sum(trivial_coeffs)

        # Test Hamiltonian was defined correctly
        assert (
            hamiltonian.terms == correct_terms
        ), f"Terms in the Hamiltonian were not constructed correctly"
        assert (
            hamiltonian.coeffs == correct_coefficients
        ), f"Coefficients in the Hamiltonian were not constructed correctly"
        assert (
            hamiltonian.constant == correct_constant
        ), f"Constant in the Hamiltonian was not correctly constructed"

        ## Exception case - number of terms does not match number of coefficients

        # Number of qubits
        n_qubits = 5

        # Define terms and coefficients
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        terms = [PauliOp("ZZ", indices) for indices in pair_indices]
        coefficients = [1 for _ in range(len(terms) + 1)]

        # Attempt construction of Hamiltonian
        with self.assertRaises(AssertionError) as context:
            hamiltonian = Hamiltonian(terms, coefficients, constant=0)

        # Check exception message
        self.assertEqual(
            "Number of Pauli terms in Hamiltonian should be same as number of coefficients",
            str(context.exception),
        )

        ## Exception case - non Pauli object term present in list of terms

        # Number of qubits
        n_qubits = 5

        # Define terms and coefficients
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        error_term = (0, 5)
        terms = [PauliOp("ZZ", indices) for indices in pair_indices] + [error_term]
        coefficients = [1 for _ in range(len(terms))]

        # Attempt construction of Hamiltonian
        with self.assertRaises(TypeError) as context:
            hamiltonian = Hamiltonian(terms, coefficients, constant=0)

        # Check exception message
        self.assertEqual(
            f"Pauli terms should be of type PauliOp and not {type(error_term)}",
            str(context.exception),
        )

    def test_hamiltonian_qureg(self):
        """
        Tests the function that generated the register in the Hamiltonian object.

        The test consists in generating registers from a set of examples.
        """

        # Number of qubits in the register
        n_qubits = 5

        # Define Pauli terms
        terms_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        terms_strings = ["XX", "YY"]
        terms = [
            PauliOp(string, indices)
            for indices in terms_indices
            for string in terms_strings
        ]

        # Define coefficients and constant
        coefficients = [1 for _ in range(len(terms))]
        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Extract physical register
        register = hamiltonian.qureg

        # Correct register
        correct_register = [0, 1, 2, 3, 4]

        # Test if the extracted register is correct
        assert np.allclose(
            register, correct_register
        ), f"Register has been incorrectly constructed"

    def test_hamiltonian_divide_into_singlets_pairs(self):
        """
        Tests the function that divides that Hamiltonian into singlets and pairs.

        The tests consists in splittig terms into singlets and pairs from a given Hamiltonian.
        """

        # Number of qubits in the register
        n_qubits = 5

        # Define Pauli terms
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        correct_linear_terms = [PauliOp("Z", indices) for indices in singlet_indices]
        correct_quadratic_terms = [PauliOp("ZZ", indices) for indices in pair_indices]

        constant_term = [PauliOp("I", (0,))]

        terms = correct_linear_terms + correct_quadratic_terms + constant_term

        # Define coefficients and constant
        correct_linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        correct_quadratic_coeffs = [1 for _ in range(len(pair_indices))]
        constant_coeff = [1]
        coefficients = correct_linear_coeffs + correct_quadratic_coeffs + constant_coeff

        constant = 1
        correct_constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Extract singlets and pairs attributes
        linear_terms = hamiltonian.qubits_singles
        quadratic_terms = hamiltonian.qubits_pairs
        linear_coeffs = hamiltonian.single_qubit_coeffs
        quadratic_coeffs = hamiltonian.pair_qubit_coeffs
        constant = hamiltonian.constant

        # Test if the hamiltonian terms and coefficients have been correctly separated
        assert (
            linear_terms == correct_linear_terms
        ), f"Linear terms have not been correctly generated"
        assert np.allclose(
            linear_coeffs, correct_linear_coeffs
        ), f"Linear coefficients have not been correctly generated"
        assert (
            quadratic_terms == correct_quadratic_terms
        ), f"Quadratic terms have not been correctly generated"
        assert np.allclose(
            quadratic_coeffs, correct_quadratic_coeffs
        ), f"Quadratic coefficients have not been correctly generated"
        assert np.allclose(
            constant, correct_constant
        ), f"Constant coefficient have not been correctly generated"

        # Exception case - generating terms beyond quadratic order

        # Define hamiltonian attributes
        terms_exc = [
            PauliOp("Z", (0,)),
            PauliOp("ZZ", (0, 1)),
            PauliOp("ZZ", (0, 2)),
            PauliOp("ZZ", (1, 2)),
            PauliOp("ZZZ", (1, 2, 3)),
        ]
        coefficients_exc = [1, 0.5, 0.5, 0.5, 3]
        constant_exc = 2

        # Test the exception is raised - Attempt construction of Hamiltonian
        with self.assertRaises(NotImplementedError) as context:
            hamiltonian_exc = Hamiltonian(terms_exc, coefficients_exc, constant_exc)

        # Check exception message
        self.assertEqual(
            "Hamiltonian only supports Linear and Quadratic terms",
            str(context.exception),
        )

    def test_hamiltonian_str(self):
        """
        Tests the function that constructs a string representation of the Hamiltonian object.

        The test consists of constructing the string representation of a given Hamiltonian.
        """

        # Number of qubits in the register
        n_qubits = 3

        # Define Pauli terms
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        linear_terms = [PauliOp("Z", indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp("ZZ", indices) for indices in pair_indices]

        terms = linear_terms + quadratic_terms

        # Define coefficients and constant
        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        coefficients = linear_coeffs + quadratic_coeffs

        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Extract string representation
        string_rep = hamiltonian.__str__()

        # Correct string representation
        correct_string_rep = "-1*Z_{0} + -1*Z_{1} + -1*Z_{2} + 0.5*Z_{0}Z_{1} + 0.5*Z_{0}Z_{2} + 0.5*Z_{1}Z_{2} + 2"

        # Test the obtained string representation
        assert (
            string_rep == correct_string_rep
        ), f"Hamiltonian string representation was incorrectly constructed"

    def test_hamiltonian_len(self):
        """
        Tests the function that computes the number of terms in the Hamiltonian.

        The test consists in checking the number of terms for a fully connected Hamiltonian
        with linear terms for an increasing number of qubits
        """

        # Initialize lists containing correct and computed lengths
        correct_lengths = []
        lengths = []

        for n_qubits in range(3, 10):
            # Compute correct lengths
            correct_lengths.append(n_qubits + n_qubits * (n_qubits - 1) / 2)

            # Define Pauli terms
            singlet_indices = [(i,) for i in range(n_qubits)]
            pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
            linear_terms = [PauliOp("Z", indices) for indices in singlet_indices]
            quadratic_terms = [PauliOp("ZZ", indices) for indices in pair_indices]

            terms = linear_terms + quadratic_terms

            # Define coefficients and constant
            linear_coeffs = [-1 for _ in range(len(singlet_indices))]
            quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
            coefficients = linear_coeffs + quadratic_coeffs

            constant = 2

            # Define Hamiltonian
            hamiltonian = Hamiltonian(terms, coefficients, constant)

            # Compute length
            length = hamiltonian.__len__()
            lengths.append(length)

        # Test that computed lengths are correct
        assert np.allclose(
            lengths, correct_lengths
        ), f"Hamiltonian lengths are incorrectly computed"

    def test_hamiltonian_expression(self):
        """
        Tests the function that computes the symbolic expression of the Hamiltonian object.

        The test consists in computing the symbolic expression for a specific example.
        """

        # Number of qubits in the register
        n_qubits = 3

        # Define Pauli terms
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        linear_terms = [PauliOp("Z", indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp("ZZ", indices) for indices in pair_indices]

        terms = linear_terms + quadratic_terms

        # Define coefficients and constant
        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        coefficients = linear_coeffs + quadratic_coeffs

        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Extract symbolic representation
        sym_rep = hamiltonian.expression

        # Correct symbolic representation
        correct_sym_rep = Symbol(str(constant))

        for ind, coeff in enumerate(linear_coeffs):
            correct_sym_rep += Symbol(str(coeff) + "Z" + "_{" + str(ind) + "}")

        for ind, coeff in enumerate(quadratic_coeffs):
            i, j = pair_indices[ind]
            correct_sym_rep += Symbol(
                str(coeff) + "Z" + "_{" + str(i) + "}" + "Z" + "_{" + str(j) + "}"
            )

        # Test if string representation was correctly generated
        assert (
            sym_rep == correct_sym_rep
        ), f"Symbolic representation was incorrectly generated"

    def test_hamiltonian_add(self):
        """
        Test the function that adds two Hamiltonian objects.

        The test consists in adding two given Hamiltonians with a repeated term.
        """

        # Number of qubits in the register
        n_qubits = 4

        # Define the first Hamiltonian
        pair_indices = [(i, j) for j in range(n_qubits - 1) for i in range(j)]
        input_terms = [PauliOp("ZZ", indices) for indices in pair_indices]

        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        input_coefficients = quadratic_coeffs
        input_constant = 2

        hamiltonian = Hamiltonian(input_terms, input_coefficients, input_constant)

        # Define the other Hamiltonian
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        linear_terms = [PauliOp("Z", indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp("ZZ", indices) for indices in pair_indices]
        other_input_terms = linear_terms + quadratic_terms

        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [
            0.5 for _ in range(int((n_qubits - 1) * (n_qubits - 2) / 2))
        ] + [1 for _ in range(n_qubits - 1)]
        other_input_coefficients = linear_coeffs + quadratic_coeffs

        other_input_constant = 0

        other_hamiltonian = Hamiltonian(
            other_input_terms, other_input_coefficients, other_input_constant
        )

        # Add hamiltonians
        hamiltonian.__add__(other_hamiltonian)

        # Exctract terms, coefficients and constants from added Hamiltonians
        terms = hamiltonian.terms
        coefficients = hamiltonian.coeffs
        constant = hamiltonian.constant

        # Create Hamiltonian dictionary - To be removed upon implementation of _simplify()
        hamiltonian_dict = {
            term.qubit_indices: coeff
            for term, coeff in zip(hamiltonian.terms, hamiltonian.coeffs)
        }

        # Correct terms, coefficients and constants
        correct_terms = input_terms + other_input_terms
        correct_coefficients = input_coefficients + other_input_coefficients
        correct_constant = input_constant + other_input_constant

        # Create Hamiltonian dictionary for correct Hamiltonian - To be removed upon implementation of _simplify()
        correct_hamiltonian_dict = {
            term.qubit_indices: coeff
            for term, coeff in zip(input_terms, input_coefficients)
        }

        # Add terms and coefficients from the other hamiltonian
        for term, coeff in zip(other_input_terms, other_input_coefficients):
            # If term was absent add it together with coefficient
            if correct_hamiltonian_dict.get(term.qubit_indices) is None:
                correct_hamiltonian_dict.update({term.qubit_indices: coeff})

            # If term was present add coefficient to current term
            else:
                correct_hamiltonian_dict[term.qubit_indices] += coeff

        # Test if Hamiltonians were added correctly
        assert (
            hamiltonian_dict == correct_hamiltonian_dict
        ), f"Hamiltonian adder did not yield correct terms and/or coefficients"
        assert np.allclose(
            constant, correct_constant
        ), f"Hamiltonian adder did not yield correct constant"

    def test_hamiltonian_squared(self):
        """
        Tests the function that computes the squared Hamiltonian, also a Hamiltonian object.

        The test consists in computing the Hamiltonian squared for two given small examples.
        """

        # First example

        # Number of qubits in the register
        n_qubits = 2

        # Define Pauli terms
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        linear_terms = [PauliOp("Z", indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp("ZZ", indices) for indices in pair_indices]
        terms = linear_terms + quadratic_terms

        # Define coefficients and constant
        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        coefficients = linear_coeffs + quadratic_coeffs
        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Compute squared hamiltonian and extract attributes
        hamiltonian_sq = hamiltonian.hamiltonian_squared

        terms_sq = hamiltonian_sq.terms
        coefficients_sq = hamiltonian_sq.coeffs
        constant_sq = hamiltonian_sq.constant

        # Correct Hamiltonian squared attributes
        correct_terms_sq = [
            PauliOp("ZZ", (0, 1)),
            PauliOp("Z", (1,)),
            PauliOp("ZZ", (0, 1)),
            PauliOp("Z", (0,)),
            PauliOp("Z", (1,)),
            PauliOp("Z", (0,)),
            PauliOp("Z", (0,)),
            PauliOp("Z", (1,)),
            PauliOp("ZZ", (0, 1)),
        ]

        correct_coefficients_sq = [1, -0.5, 1, -0.5, -0.5, -0.5, -4, -4, 2]

        correct_constant_sq = (
            4 + n_qubits * (-1) ** 2 + n_qubits * (n_qubits - 1) / 2 * (0.5) ** 2
        )

        # Test if Hamiltonian squared was correctly computed
        assert (
            terms_sq == correct_terms_sq
        ), f"Hamiltonian squared did not yield correct terms"
        assert np.allclose(
            coefficients_sq, correct_coefficients_sq
        ), f"Hamiltonian squared did not yield correct coefficients"
        assert np.allclose(
            constant_sq, correct_constant_sq
        ), f"Hamiltonian squared did not yield correct constant"

        # Second example

        # Number of qubits in the register
        n_qubits = 2

        # Define Pauli terms
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        terms_strings = ["XX", "YY"]
        terms = [
            PauliOp(string, indices)
            for indices in pair_indices
            for string in terms_strings
        ]

        # Define coefficients and constant
        coefficients = [1 for _ in range(len(terms))]
        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Compute squared hamiltonian and extract attributes
        hamiltonian_sq = hamiltonian.hamiltonian_squared

        terms_sq = hamiltonian_sq.terms
        coefficients_sq = hamiltonian_sq.coeffs
        constant_sq = hamiltonian_sq.constant

        # Correct Hamiltonian squared attributes
        correct_terms_sq = [
            PauliOp("ZZ", (0, 1)),
            PauliOp("ZZ", (0, 1)),
            PauliOp("XX", (0, 1)),
            PauliOp("YY", (0, 1)),
        ]
        correct_coefficients_sq = [-1, -1, 4, 4]
        correct_constant_sq = 6

        # Test if Hamiltonian squared was correctly computed
        assert (
            terms_sq == correct_terms_sq
        ), f"Hamiltonian squared did not yield correct terms"
        assert np.allclose(
            coefficients_sq, correct_coefficients_sq
        ), f"Hamiltonian squared did not yield correct coefficients"
        assert np.allclose(
            constant_sq, correct_constant_sq
        ), f"Hamiltonian squared did not yield correct constant"

    def test_classical_hamiltonian(self):
        """
        Tests the function that generates a classical Hamiltonian, i.e. a Hamiltonian composed only
        of Z Pauli operators.

        The tests consists in generating a Hamiltonian from a given set of graph edges and weights.
        """

        # Number of nodes in the graph, i.e. qubits in the register
        n_nodes = 5

        # Graph properties
        input_edges = [(i,) for i in range(n_nodes)] + [
            (i, j) for j in range(n_nodes) for i in range(j)
        ]
        input_weights = [-1 for _ in range(len(input_edges))]
        input_constant = 2

        # Generate classical Hamiltonian and extract attributes
        cl_hamiltonian = Hamiltonian.classical_hamiltonian(
            input_edges, input_weights, input_constant
        )

        terms = cl_hamiltonian.terms
        terms_indices = [term.qubit_indices for term in terms]
        coefficients = cl_hamiltonian.coeffs
        constant = cl_hamiltonian.constant

        # Test that the generated Hamiltonian match the properties of the input graph
        assert (
            terms_indices == input_edges
        ), f"Classical Hamiltonian terms did not match input graph edges"
        assert np.allclose(
            coefficients, input_weights
        ), f"Hamiltonian squared did not yield correct coefficients"
        assert np.allclose(
            constant, input_constant
        ), f"Hamiltonian squared did not yield correct constant"

        # Exception case - inserting coefficients that are not interegers or floats

        # Define hamiltonian attributes
        terms_exc = [(0, 1), (0, 2), (1, 2)]
        constant_exc = 2

        # Define different types of coefficients containing exceptions
        coefficients_exc_types = [[-1j, 1, -1j], [2, 3, 0.1 * 1j]]

        # Check for every coefficient type
        for coefficients_exc in coefficients_exc_types:
            # Test the exception is raised - Attempt construction of Hamiltonian
            with self.assertRaises(ValueError) as context:
                hamiltonian_exc = Hamiltonian.classical_hamiltonian(
                    terms_exc, coefficients_exc, constant_exc
                )

        # Check exception message
        self.assertEqual(
            "Classical Hamiltonians only support Integer or Float coefficients",
            str(context.exception),
        )

        # Exception case - generating terms beyond quadratic order

        # Define hamiltonian attributes
        terms_exc = [(0,), (0, 1), (0, 2), (1, 2), (1, 2, 3)]
        coefficients_exc = [1, 0.5, 0.5, 0.5, 3]
        constant_exc = 2

        # Test the exception is raised - Attempt construction of Hamiltonian
        with self.assertRaises(ValueError) as context:
            hamiltonian_exc = Hamiltonian.classical_hamiltonian(
                terms_exc, coefficients_exc, constant_exc
            )

        # Check exception message
        self.assertEqual(
            "Hamiltonian only supports Linear and Quadratic terms",
            str(context.exception),
        )

    def test_hamiltonian_dict(self):
        """Test if the method produces
        the correct dictionary corresponding to the Hamiltonian.

        It creates a dictionary of the form
        {(i, j): coeff} where i, j are the indices of the qubits if
        argument `classical` is set true in the function otherwise,
        produces a dictionary of the form {PauliOp}: coeff}
        """

        # TEST 1
        # Define Hamiltonian attributes
        terms = [[0, 1], [0, 2], [1, 2]]
        coefficients = [1, 0.5, 0.5]
        constant = 2

        # Generate Hamiltonian
        hamiltonian = Hamiltonian.classical_hamiltonian(terms, coefficients, constant)

        # Generate dictionary
        hamiltonian_dict = hamiltonian.hamiltonian_dict()

        # Correct dictionary
        correct_dict = {(0, 1): 1, (0, 2): 0.5, (1, 2): 0.5, (): 2}

        # Test if dictionary was correctly computed
        assert (
            hamiltonian_dict == correct_dict
        ), f"Hamiltonian dictionary did not yield correct dictionary"

        # TEST 2
        terms = [PauliOp("XX", (0, 1)), PauliOp("YY", (0, 2))]
        coeffs = [1, 0.5]
        constant = 0

        # Generate Hamiltonian
        hamiltonian = Hamiltonian(terms, coeffs, constant)

        # Generate dictionaries
        hamiltonian_dict_indices = hamiltonian.hamiltonian_dict(classical=True)
        hamiltonian_dict_pauliop = hamiltonian.hamiltonian_dict(classical=False)

        # Correct dictionaries
        correct_dict_indices = {(0, 1): 1, (0, 2): 0.5, (): 0}
        correct_dict_pauliop = {
            PauliOp("XX", (0, 1)): 1,
            PauliOp("YY", (0, 2)): 0.5,
            (): 0,
        }

        assert (
            hamiltonian_dict_indices == correct_dict_indices
        ), "Hamiltonian dictionary did not yield correct dictionary"

        assert (
            hamiltonian_dict_pauliop == correct_dict_pauliop
        ), "Hamiltonian dictionary did not yield correct dictionary"

    def test_hamiltonian_as_matrix(self):
        """Test if the method produces
        the correct matrix corresponding to given Hamiltonian.

        """
        P_list = ["XIIZ", "YIII", "IZXI", "IXIY", "IIIZ", "IIXI", "ZZII"]
        coeff_list = list(range(1, 8))

        Pauli_list = []
        for P in P_list:
            P = np.array(list(P))
            non_I = np.where(P != "I")[0]
            Pauli_list.append(PauliOp(P[non_I], tuple(non_I)))

        H = Hamiltonian(Pauli_list, coeff_list, 0)

        solution = np.array(
            [
                [
                    12.0 + 0.0j,
                    0.0 + 0.0j,
                    9.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 - 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    1.0 - 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    2.0 + 0.0j,
                    0.0 + 0.0j,
                    9.0 + 0.0j,
                    0.0 + 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -1.0 - 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    9.0 + 0.0j,
                    0.0 + 0.0j,
                    12.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 - 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    1.0 - 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    9.0 + 0.0j,
                    0.0 + 0.0j,
                    2.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -1.0 - 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 - 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -2.0 + 0.0j,
                    0.0 + 0.0j,
                    3.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    1.0 - 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -12.0 + 0.0j,
                    0.0 + 0.0j,
                    3.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -1.0 - 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 - 4.0j,
                    3.0 + 0.0j,
                    0.0 + 0.0j,
                    -2.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    1.0 - 2.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    3.0 + 0.0j,
                    0.0 + 0.0j,
                    -12.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -1.0 - 2.0j,
                ],
                [
                    1.0 + 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -2.0 + 0.0j,
                    0.0 + 0.0j,
                    9.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 - 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    -1.0 + 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -12.0 + 0.0j,
                    0.0 + 0.0j,
                    9.0 + 0.0j,
                    0.0 + 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    1.0 + 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    9.0 + 0.0j,
                    0.0 + 0.0j,
                    -2.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 - 4.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -1.0 + 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    9.0 + 0.0j,
                    0.0 + 0.0j,
                    -12.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 4.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    1.0 + 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 - 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    12.0 + 0.0j,
                    0.0 + 0.0j,
                    3.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -1.0 + 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    2.0 + 0.0j,
                    0.0 + 0.0j,
                    3.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    1.0 + 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 - 4.0j,
                    3.0 + 0.0j,
                    0.0 + 0.0j,
                    12.0 + 0.0j,
                    0.0 + 0.0j,
                ],
                [
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    -1.0 + 2.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    0.0 + 4.0j,
                    0.0 + 0.0j,
                    0.0 + 0.0j,
                    3.0 + 0.0j,
                    0.0 + 0.0j,
                    2.0 + 0.0j,
                ],
            ]
        )

        assert np.allclose(
            solution, H.as_matrix.toarray()
        ), "as_matrix method not correct"


if __name__ == "__main__":
    unittest.main()



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
A set of functions to implement pennylane optimization algorithms.
Read https://docs.pennylane.ai/en/stable/introduction/interfaces.html#optimizers
Optimisers requiring a pennylane backend haven't been implemented yet.
Similarly as with the custom optimization methods Scipy `minimize` is used. Extends available scipy methods.
"""

from openqaoa.optimizers import pennylane as pl
import inspect
from scipy.optimize import OptimizeResult
import numpy as np

AVAILABLE_OPTIMIZERS = {  # optimizers implemented
    "pennylane_adagrad": pl.AdagradOptimizer,
    "pennylane_adam": pl.AdamOptimizer,
    "pennylane_vgd": pl.GradientDescentOptimizer,
    "pennylane_momentum": pl.MomentumOptimizer,
    "pennylane_nesterov_momentum": pl.NesterovMomentumOptimizer,
    "pennylane_rmsprop": pl.RMSPropOptimizer,
    "pennylane_rotosolve": pl.RotosolveOptimizer,
    "pennylane_spsa": pl.SPSAOptimizer,
}


def pennylane_optimizer(
    fun,
    x0,
    args=(),
    maxfev=None,
    pennylane_method="vgd",
    maxiter=100,
    tol=10 ** (-6),
    jac=None,
    callback=None,
    nums_frequency=None,
    spectra=None,
    shifts=None,
    **options
):
    """
    Minimize a function `fun` using some pennylane method.
    To check available methods look at the available_methods_dict variable.
    Read https://docs.pennylane.ai/en/stable/introduction/interfaces.html#optimizers

    PARAMETERS
    ----------
    fun : callable
        Function to minimize
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    pennylane_method : string, optional
        Optimizer method to compute the steps.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than the difference between two steps, terminate optimization.
    jac : callable, optinal
        Callable gradient function. Required for all methods but rotosolve and spsa.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.
    options : dict, optional
        Dictionary where keys are the arguments for the optimizers object, and
        the values are the values to pass to these arguments.
        To know all the possible options see https://docs.pennylane.ai/en/stable/introduction/interfaces.html#optimizers.
    nums_frequency : dict[dict], optional
        It is required for rotosolve method
        The number of frequencies in the fun per parameter.
    spectra : dict[dict], optional
        It is required for rotosolve method
        Frequency spectra in the objective_fn per parameter.
    shifts : dict[dict], optional
        It is required for rotosolve method
        Shift angles for the reconstruction per parameter.
        Read https://docs.pennylane.ai/en/stable/code/api/pennylane.RotosolveOptimizer.html#pennylane.RotosolveOptimizer.step for more information.


    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    """

    def cost(
        params, **k
    ):  # define a function to convert the params list from pennylane to numpy
        return fun(np.array(params), *k)

    optimizer = AVAILABLE_OPTIMIZERS[pennylane_method]  # define the optimizer

    # get optimizer arguments
    arguments = inspect.signature(optimizer).parameters.keys()
    options_keys = list(options.keys())

    # check which values of the options dict can be passed to the optimizer (pop the others)
    for key in options_keys:
        if key not in arguments:
            options.pop(key)
        if "maxiter" in arguments:
            options["maxiter"] = maxiter

    optimizer = optimizer(**options)  # pass the arguments

    bestx = pl.numpy.array(x0, requires_grad=True)
    besty = cost(x0, *args)
    funcalls = 1  # tracks no. of function evals.
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(besty)
    while improved and not stop and niter < maxiter:
        improved = False

        # compute step (depends on the optimizer)
        if pennylane_method in [
            "pennylane_adagrad",
            "pennylane_adam",
            "pennylane_vgd",
            "pennylane_momentum",
            "pennylane_nesterov_momentum",
            "pennylane_rmsprop",
        ]:
            testx, testy = optimizer.step_and_cost(cost, bestx, *args, grad_fn=jac)
        elif pennylane_method in ["pennylane_rotosolve"]:
            testx, testy = optimizer.step_and_cost(
                cost,
                bestx,
                *args,
                nums_frequency={"params": {(i,): 1 for i in range(bestx.size)}}
                if not nums_frequency
                else nums_frequency,
                spectra=spectra,
                shifts=shifts,
                full_output=False,
            )
        elif pennylane_method in ["pennylane_spsa"]:
            testx, testy = optimizer.step_and_cost(cost, bestx, *args)

        # check if stable
        if np.abs(besty - testy) < tol and niter > 1:
            improved = False

        else:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

        niter += 1

    return OptimizeResult(
        fun=besty, x=np.array(bestx), nit=niter, nfev=funcalls, success=(niter > 1)
    )



"""
Conversion functions for the different QAOA Parametrizations. So far we only
only support going from less to more specialiced parametrizations. The type
tree looks as follows:
   Extended   <--------- FourierExtended
       ^                      ^
       |                      |
StandardWithBias <------ FourierWithBias
       ^                      ^
       |                      |
    Standard  <----------- Fourier
       ^
       |
    Annealing
"""
from __future__ import annotations
from copy import deepcopy
import numpy as np
from scipy.fftpack import dct, dst

from . import annealingparams, standardparams, fourierparams, extendedparams

QAOAVariationalAnnealingParams = annealingparams.QAOAVariationalAnnealingParams
QAOAVariationalStandardParams = standardparams.QAOAVariationalStandardParams
QAOAVariationalExtendedParams = extendedparams.QAOAVariationalExtendedParams
QAOAVariationalStandardWithBiasParams = (
    standardparams.QAOAVariationalStandardWithBiasParams
)
QAOAVariationalFourierParams = fourierparams.QAOAVariationalFourierParams
QAOAVariationalFourierWithBiasParams = (
    fourierparams.QAOAVariationalFourierWithBiasParams
)
QAOAVariationalFourierExtendedParams = (
    fourierparams.QAOAVariationalFourierExtendedParams
)


def annealing_to_standard(
    params: QAOAVariationalAnnealingParams,
) -> QAOAVariationalStandardParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalStandardParams
    out.betas = params._annealing_time * (1 - params.schedule) / params.p
    out.gammas = params._annealing_time * params.schedule / params.p

    # and clean up after us
    del out.__schedule
    del out._annealing_time

    return out


def standard_to_standard_w_bias(
    params: QAOAVariationalStandardParams,
) -> QAOAVariationalStandardWithBiasParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalStandardWithBiasParams
    out.gammas_singles = params.gammas
    out.gammas_pairs = params.gammas

    # and clean up after us
    del out.__gammas

    return out


def standard_w_bias_to_extended(
    params: QAOAVariationalStandardWithBiasParams,
) -> QAOAVariationalExtendedParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalExtendedParams
    out.betas_singles = np.outer(params.betas, np.ones(len(params.mixer_1q_coeffs)))
    out.betas_pairs = np.outer(params.betas, np.ones(len(params.mixer_2q_coeffs)))

    out.gammas_singles = np.outer(
        params.gammas_singles, np.ones(len(params.cost_1q_coeffs))
    )
    out.gammas_pairs = np.outer(
        params.gammas_pairs, np.ones(len(params.cost_2q_coeffs))
    )
    return out


def fourier_to_standard(
    params: QAOAVariationalFourierParams,
) -> QAOAVariationalStandardParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalStandardParams
    out.betas = dct(params.v, n=out.p)
    out.gammas = dst(params.u, n=out.p)

    # and clean up
    del out.__u
    del out.__v
    del out.q

    return out


def fourier_w_bias_to_standard_w_bias(
    params: QAOAVariationalFourierWithBiasParams,
) -> QAOAVariationalStandardWithBiasParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalStandardWithBiasParams
    out.betas = dct(params.v, n=out.p)
    out.gammas_singles = dst(params.u_singles, n=out.p)
    out.gammas_pairs = dst(params.u_pairs, n=out.p)

    # and clean up
    del out.__u_singles
    del out.__u_pairs
    del out.__v
    del out.q

    return out


def fourier_to_fourier_w_bias(
    params: QAOAVariationalFourierParams,
) -> QAOAVariationalFourierWithBiasParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalFourierWithBiasParams
    out.u_singles = params.u
    out.u_pairs = params.u

    # and clean up
    del out.__u

    return out


def fourier_w_bias_to_fourier_extended(
    params: QAOAVariationalFourierWithBiasParams,
) -> QAOAVariationalFourierExtendedParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalFourierExtendedParams
    out.v_singles = np.outer(params.v, np.ones(len(params.mixer_1q_coeffs)))
    out.v_pairs = np.outer(params.v, np.ones(len(params.mixer_2q_coeffs)))
    out.u_singles = np.outer(params.u_singles, np.ones(len(params.cost_1q_coeffs)))
    out.u_pairs = np.outer(params.u_pairs, np.ones(len(params.cost_2q_coeffs)))

    del out.__v

    return out


def fourier_extended_to_extended(
    params: QAOAVariationalFourierExtendedParams,
) -> QAOAVariationalExtendedParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalExtendedParams
    out.betas_singles = dct(params.v_singles, n=out.p, axis=0)
    out.betas_pairs = dct(params.v_pairs, n=out.p, axis=0)
    out.gammas_singles = dst(params.u_singles, n=out.p, axis=0)
    out.gammas_pairs = dst(params.u_pairs, n=out.p, axis=0)

    # and clean up
    del out.__u_singles
    del out.__u_pairs
    del out.__v_singles
    del out.__v_pairs
    del out.q

    return out


# #############################################################################
# And now all the possible compositions as well:
# Todo: Create this code automatically by traversing the tree?
# #############################################################################


def annealing_to_standard_w_bias(
    params: QAOAVariationalAnnealingParams,
) -> QAOAVariationalStandardWithBiasParams:
    return standard_to_standard_w_bias(annealing_to_standard(params))


def annealing_to_extended(
    params: QAOAVariationalAnnealingParams,
) -> QAOAVariationalExtendedParams:
    return standard_w_bias_to_extended(annealing_to_standard_w_bias(params))


def standard_to_extended(
    params: QAOAVariationalStandardParams,
) -> QAOAVariationalExtendedParams:
    return standard_w_bias_to_extended(standard_to_standard_w_bias(params))


def fourier_to_fourier_extended(
    params: QAOAVariationalFourierParams,
) -> QAOAVariationalFourierExtendedParams:
    return fourier_w_bias_to_fourier_extended(fourier_to_fourier_w_bias(params))


def fourier_to_standard_w_bias(
    params: QAOAVariationalFourierParams,
) -> QAOAVariationalStandardWithBiasParams:
    return standard_to_standard_w_bias(fourier_to_standard(params))


def fourier_to_extended(
    params: QAOAVariationalFourierParams,
) -> QAOAVariationalExtendedParams:
    return standard_w_bias_to_extended(fourier_to_standard_w_bias(params))


def fourier_w_bias_to_extended(
    params: QAOAVariationalFourierWithBiasParams,
) -> QAOAVariationalExtendedParams:
    return standard_w_bias_to_extended(fourier_w_bias_to_standard_w_bias(params))


# A dict with all the conversion functions accessible by
# (input_type, output_type)
conversion_functions = {
    (
        QAOAVariationalFourierExtendedParams,
        QAOAVariationalExtendedParams,
    ): fourier_extended_to_extended,
    (
        QAOAVariationalFourierWithBiasParams,
        QAOAVariationalStandardWithBiasParams,
    ): fourier_w_bias_to_standard_w_bias,
    (QAOAVariationalFourierParams, QAOAVariationalStandardParams): fourier_to_standard,
    (
        QAOAVariationalAnnealingParams,
        QAOAVariationalStandardParams,
    ): annealing_to_standard,
    (
        QAOAVariationalStandardParams,
        QAOAVariationalStandardWithBiasParams,
    ): standard_to_standard_w_bias,
    (
        QAOAVariationalStandardWithBiasParams,
        QAOAVariationalExtendedParams,
    ): standard_w_bias_to_extended,
    (
        QAOAVariationalFourierParams,
        QAOAVariationalFourierWithBiasParams,
    ): fourier_to_fourier_w_bias,
    (
        QAOAVariationalFourierWithBiasParams,
        QAOAVariationalFourierExtendedParams,
    ): fourier_w_bias_to_fourier_extended,
    (
        QAOAVariationalAnnealingParams,
        QAOAVariationalStandardWithBiasParams,
    ): annealing_to_standard_w_bias,
    (
        QAOAVariationalAnnealingParams,
        QAOAVariationalExtendedParams,
    ): annealing_to_extended,
    (
        QAOAVariationalStandardParams,
        QAOAVariationalExtendedParams,
    ): standard_to_extended,
    (
        QAOAVariationalFourierParams,
        QAOAVariationalFourierExtendedParams,
    ): fourier_to_fourier_extended,
    (
        QAOAVariationalFourierParams,
        QAOAVariationalStandardWithBiasParams,
    ): fourier_to_standard_w_bias,
    (QAOAVariationalFourierParams, QAOAVariationalExtendedParams): fourier_to_extended,
    (
        QAOAVariationalFourierWithBiasParams,
        QAOAVariationalExtendedParams,
    ): fourier_w_bias_to_extended,
}


def converter(params, target_type: type):
    """
    Convert ``params`` to type ``target_type``
    Parameters
    ----------
    params:
        The input parameters
    target_type:
        The target type
    Returns
    -------
    target_type:
        The converted parameters
    Raises
    ------
    TypeError:
        If conversion from type(params) to target_type is not supported.
    """
    try:
        out = conversion_functions[(type(params), target_type)](params)
        return out
    except KeyError:
        raise TypeError(
            f"Conversion from {type(params)} to {target_type} " "not supported."
        )



import numpy as np
from scipy.optimize import OptimizeResult

#### Experimental ############################################################


def SPSA(
    fun,
    x0,
    args=(),
    maxfev=None,
    a0=0.01,
    c0=0.01,
    A=1,
    alpha=0.602,
    gamma=0.101,
    maxiter=100,
    tol=10 ** (-6),
    jac=None,
    callback=None,
    **options
):
    """
    Minimize a function `fun` using the Simultaneous Perturbation Stochastic Approximation (SPSA) method.
    scipy.optimize.minimize compatible implementation of SPSA for `method` == 'spsa'.
    Refer to https://www.jhuapl.edu/spsa/PDF-SPSA/Spall_Implementation_of_the_Simultaneous.PDF
    for in depth explanation of parameters and optimal parameter choices.

    PARAMETERS
    ----------
    fun :
        Function to minimize.
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    a0 : float
        Initial stepsize.
    c0 : float
        Initial finite difference.
    A : float
        Initial decay constant.
    alpha : float
        Decay rate parameter for `a`.
    gamma : float
        Decay rate parameter for `c`.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates;
        if `tol` is larger than the difference between
        two steps, terminate optimization.
    jac : callable
        Callable gradient function.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.

    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    """

    def grad_SPSA(params, c):
        delta = 2 * np.random.randint(0, 2, size=len(params)) - 1
        return np.real(
            (fun(params + c * delta) - fun(params - c * delta)) * delta / (2 * c)
        )

    bestx = x0
    besty = fun(x0)
    funcalls = 1  # tracks no. of function evals.
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(fun(testx, *args))

    a = a0
    c = c0
    while improved and not stop and niter < maxiter:
        improved = False
        niter += 1

        # gain sequences
        a = a / (A + niter + 1) ** alpha
        c = c / (niter + 1) ** gamma

        # compute gradient descent step
        testx = testx - a * grad_SPSA(testx, c)
        testy = np.real(fun(testx, *args))

        if np.abs(besty - testy) < tol:
            improved = False

        else:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

    return OptimizeResult(
        fun=besty, x=bestx, nit=niter, nfev=funcalls, success=(niter > 1)
    )



from setuptools import setup, find_packages
from os import getcwd

current_path = getcwd()

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("openqaoa/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

requirements = open("requirements.txt").readlines()
requirements = [r.strip() for r in requirements]

requirements_docs = open("requirements_docs.txt").readlines()
requirements_docs = [r.strip() for r in requirements_docs]

requirements_test = open("requirements_test.txt").readlines()
requirements_test = [r.strip() for r in requirements_test]

setup(
    name="openqaoa-core",
    python_requires=">=3.8, <3.11",
    version=version,
    author="Entropica Labs",
    packages=find_packages(where="."),
    entry_points={"openqaoa.plugins": []},
    url="https://github.com/entropicalabs/openqaoa",
    install_requires=requirements,
    license="MIT",
    description="OpenQAOA is a python open-source multi-backend Software Development Kit to create, customise and execute the Quantum Approximate Optimisation Algorithm (QAOA) on Noisy Intermediate-Scale Quantum (NISQ) devices, and simulators",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    keywords="quantum optimisation SDK",
    extras_require={
        "docs": requirements_docs,
        "tests": requirements_test,
        "all": requirements_docs + requirements_test,
    },
)



import numpy as np
from math import log2
from random import shuffle
from typing import List, Union
from matplotlib import pyplot as plt
from copy import deepcopy
from IPython.display import clear_output
import time

from . import QAOA
from ...backends import create_device


class QAOABenchmark:
    """
    Benchmark is a class that implements benchmarking for QAOA.

    Attributes
    ----------
    qaoa : QAOA
        The QAOA object to be benchmarked.
    reference : QAOA
        The reference QAOA object, which will be used to compare the results of the benchmarked QAOA object.
    ranges : List[tuple]
        The ranges of the variate parameters of the benchmarked QAOA object.
    ranges_reference : List[tuple]
        The ranges of the variate parameters of the reference QAOA object.
    values : np.ndarray
        The values of the benchmarked QAOA object.
    values_reference : np.ndarray
        The values of the reference QAOA object.
    difference : np.ndarray
        The difference between the values of the benchmarked QAOA object and the values of the reference QAOA object.
    difference_mean : float
        The mean of the difference between the values of the benchmarked QAOA object and the values of the reference QAOA object.
    """

    def __init__(self, qaoa: QAOA):
        """
        Constructor for the Benchmark class.

        Parameters
        ----------
        qaoa : QAOA
            The QAOA object to be benchmarked.
        """
        # check if the QAOA object inputted is valid and save it
        assert isinstance(qaoa, QAOA), "`qaoa` must be an instance of QAOA"
        assert qaoa.compiled, "`qaoa` must be compiled before benchmarking"
        self.qaoa = qaoa

        # create a reference QAOA object, which will be used to compare the results of the benchmarked QAOA object
        self.reference = QAOA()
        self.reference.circuit_properties = (
            self.qaoa.circuit_properties
        )  # copy the circuit properties of the benchmarked QAOA object
        try:
            self.reference.set_device(
                create_device(location="local", name="analytical_simulator")
            )
            self.reference.compile(self.qaoa.problem)
        except:
            self.reference.set_device(
                create_device(location="local", name="vectorized")
            )
            self.reference.compile(self.qaoa.problem)
            print(
                "Warning: vectorized simulator will be used for the reference, since the analytical simulator is not available for the circuit properties of the benchmarked QAOA object"
            )

        # initialize values and ranges
        self.values = None
        self.values_reference = None
        self.ranges = None
        self.ranges_reference = None

    @property
    def difference(self):
        "The difference between the values of the benchmarked QAOA object and the values of the reference QAOA object."
        if self.values is None:
            raise Exception(
                "You must run the benchmark before calculating the difference"
            )
        if self.values_reference is None:
            raise Exception(
                "You must run the reference before calculating the difference"
            )
        if self.values.shape != self.values_reference.shape:
            raise Exception(
                "The ranges and number of points of the values and reference values must be the same"
            )
        return self.values - self.values_reference

    @property
    def difference_mean(self):
        "The mean of the difference between the values of the benchmarked QAOA object and the values of the reference QAOA object."
        return np.mean(self.difference)

    @property
    def difference_std(self):
        "The standard deviation of the difference between the values of the benchmarked QAOA object and the values of the reference QAOA object."
        return np.std(self.difference)

    def __assert_run_inputs(
        self,
        n_points_axis: int,
        ranges: List[tuple],
        run_main: bool,
        run_reference: bool,
        plot: bool,
        plot_difference: bool,
        plot_options: dict,
        verbose: bool,
    ):
        "Private method that checks the inputs of the run (and run_reference) method."
        assert isinstance(
            n_points_axis, int
        ), "The number of points per axis must be an integer"
        assert (
            isinstance(ranges, list) or ranges is None
        ), "The ranges argument must be a list of tuples: (min, max) or (value,)"
        assert len(ranges) == len(
            self.qaoa.variate_params
        ), "The number of ranges must be equal to the number of variate parameters, which is {}".format(
            len(self.qaoa.variate_params)
        )
        assert all(
            [isinstance(r, tuple) or isinstance(r, list) for r in ranges]
        ), "Each range must be a tuple: (min, max) or (value,)"
        assert all(
            [len(r) == 1 or len(r) == 2 for r in ranges]
        ), "Each range must be a tuple of length 1 or 2: (min, max) or (value,)"
        assert (
            len([r for r in ranges if len(r) == 2]) > 0
        ), "At least one range must be a tuple of length 2: (min, max)"
        for bools in [
            "run_main",
            "run_reference",
            "plot",
            "plot_difference",
            "verbose",
        ]:
            assert isinstance(
                eval(bools), bool
            ), "The {} argument must be a boolean".format(bools)
        assert (
            run_main or run_reference
        ), "You must run the main or the reference or both"
        assert isinstance(
            plot_options, dict
        ), "The plot_options argument must be a dictionary"

    def __print_expected_time(self, k, n_points):
        "Private method that prints the expected remaining time to complete the benchmark."
        print("\rPoint {} out of {}.".format(k + 1, n_points), end="")
        if k == 0:
            self.__start_time = time.time()
        elif k > 0:
            expected_time = (time.time() - self.__start_time) * (n_points - k) / k
            print(
                " Expected remaining time to complete: {}, it will be finished at {}.".format(
                    time.strftime("%H:%M:%S", time.localtime(expected_time)),
                    time.strftime(
                        "%H:%M:%S", time.localtime(time.time() + expected_time)
                    ),
                ),
                end="",
            )

    def run(
        self,
        n_points_axis,
        ranges: List[tuple],
        run_main: bool = True,
        run_reference: bool = True,
        plot: bool = False,
        plot_every: int = 1000,
        plot_difference: bool = False,
        plot_options: dict = {},
        verbose: bool = True,
    ):
        """
        Evaluates the QAOA circuit of the benchmarked (and the reference) QAOA object for a given number of points per axis and ranges.

        Parameters
        ----------
        n_points_axis : int
            The number of points per axis. It is recommended to use a number that is a power of 2 + 1: 2**k+1 (where, k is an integer).
            Using this number will ensure that the points evaluated are ordered such that there are k rounds, each round covers the whole range,
            and each round has the double number of points (per axis) of the previous round.
        ranges : List[tuple]
            The sweep ranges of the parameters. The expected format is a list of tuples: (min, max) or (value,). One tuple per parameter.
            If the length of the tuple is 1, the parameter will be fixed to the value of the tuple.
            If the length of the tuple is 2, the parameter will be swept from the first value of the tuple to the second value of the tuple.
        run_main : bool, optional
            If True, the benchmarked QAOA object will be evaluated for the requested points.
            The default is True.
        run_reference : bool, optional
            If True, the reference QAOA object will be evaluated for the same points.
            The default is True.
        plot : bool, optional
            If True, the values will be plotted.
            The default is False.
        plot_every : int, optional
            The number of evaluations after which the plot will be updated.
            The default is 1000.
        plot_difference : bool, optional
            If True, the difference between the values of the benchmarked QAOA object and the values of the reference QAOA object will be plotted.
            The default is False.
        plot_options : dict, optional
            The options for the plot. The expected format is a dictionary with the keys of the plot method of this class.
            The default is {}.
        verbose : bool, optional
            If True, the expected remaining time to complete the benchmark will be printed.
            The default is True.
        """

        # check the inputs
        self.__assert_run_inputs(
            n_points_axis,
            ranges,
            run_main,
            run_reference,
            plot,
            plot_difference,
            plot_options,
            verbose,
        )

        # plot options
        plot_options = {**{"verbose": verbose}, **plot_options}

        # save the ranges
        if run_main:
            self.ranges = ranges
        if run_reference:
            self.ranges_reference = ranges

        # get only the ranges that will be swept, and the number of parameters swept
        ranges_to_use = [r for r in ranges if len(r) == 2]
        n_params = len(ranges_to_use)

        # prepare the axes of the grid
        axes = [
            np.linspace(r[0], r[1], n_points_axis) for r in ranges_to_use
        ]  # the starting and ending points are given by the ranges and the number of points is given by n_points_axis

        # function that given the values used for the parameters swept, returns the values of all the parameters in order
        params = lambda new_params: [
            r[0] if len(r) == 1 else new_params.pop(0) for r in ranges
        ]

        # copy the parameters object, which will be updated before each evaluation
        params_obj = deepcopy(self.qaoa.variate_params)

        # initialize the array that will contain the values of the benchmarked QAOA object
        if run_main:
            self.values = np.zeros([n_points_axis for _ in range(n_params)])
        if run_reference:
            self.values_reference = np.zeros([n_points_axis for _ in range(n_params)])

        # initialize the arrays that will contain the difference between the values of the benchmarked QAOA object and the values of the reference QAOA object
        both_qaoa = [
            qaoa
            for qaoa, boolean in zip(
                [self.qaoa, self.reference], [run_main, run_reference]
            )
            if boolean
        ]
        both_values = [
            values
            for values, boolean in zip(
                [self.values, self.values_reference], [run_main, run_reference]
            )
            if boolean
        ]
        both_strings = [
            string
            for string, boolean in zip(
                ["benchmark", "reference"], [run_main, run_reference]
            )
            if boolean
        ]

        # loop over the benchmarked QAOA object and the reference QAOA object (if requested)
        for qaoa, values, string in zip(both_qaoa, both_values, both_strings):
            if verbose:
                print(f"Running {string}.")

            # evaluate all the points in the grid, in the order provided by the function __ordered_points. We loop over the indices of the grid.
            for k, i_point in enumerate(self.__ordered_points(n_params, n_points_axis)):
                if verbose:
                    self.__print_expected_time(
                        k, n_points_axis**n_params
                    )  # print the expected remaining time, info for the user

                new_params = [
                    axis[i] for axis, i in zip(axes, i_point)
                ]  # from the indices of the grid, get the values of the parameters that will be evaluated
                params_obj.update_from_raw(params(new_params))
                values[tuple(i_point[::-1])] = qaoa.backend.expectation(params_obj)

                # plot every 'plot_every' points and at the end
                if (
                    (
                        (k % plot_every == 0 and k > 0)
                        or k + 1 == n_points_axis**n_params
                    )
                    and plot
                    and string == "benchmark"
                ):
                    clear_output(wait=True)

                    if (not k + 1 == n_points_axis**n_params) and n_params == 1:
                        plot_opt = {"marker": "o", "linestyle": "", "markersize": 3}
                    else:
                        plot_opt = {}
                    self.plot(plot_options=plot_opt, **plot_options)

            if verbose:
                print(
                    " "
                )  # print a blank line, necessary because previous print: end=""

        # plot the reference if requested
        if plot and run_reference:
            self.plot(main=False, reference=True, difference=False, **plot_options)

        # plot the difference if requested and the reference has been run
        if plot_difference:
            try:
                self.plot(main=False, reference=False, difference=True, **plot_options)
            except:
                print(
                    "The difference cannot be plotted because the reference has not been run."
                )

    def plot(
        self,
        ax: plt.Axes = None,
        title: Union[str, List[str]] = None,
        labels: List[str] = None,
        labels_legend: Union[str, List[str]] = None,
        main: bool = True,
        reference: bool = False,
        difference: bool = False,
        one_plot: bool = False,
        # params_to_plot:List[int] = None, TODO: difficult because you need to specify the values of the other parameters
        plot_options: dict = {},
        verbose: bool = True,
    ):
        """
        Plots the values of the benchmarked QAOA object and/or the values of the reference QAOA
          object and/or the difference between the values of the benchmarked QAOA object and the values of the reference QAOA object.

        Parameters
        ----------
        ax : plt.Axes, optional
            The matplotlib Axes object to plot on. If None, a new figure will be created.
            The default is None.
        title : Union[str, List[str]], optional
            The title of the plot. The expected format is a string or a list of strings: one for each plot.
            If None, the title will be the default: `main plot`, `reference plot` or `difference plot`.
            The default is None.
        labels : List[str], optional
            The labels of the axes. The expected format is a list of two strings: one for each axis.
            If None, the labels will be the number of the parameters.
            The default is None.
        labels_legend : Union[str, List[str]], optional
            The labels of the legend. The expected format is a string or a list of strings: one for each line.
            It is only available if the sweep is over just one parameter, and if `one_plot` is True or `ax` is not None.
            If None, the labels will be the default: `main`, `reference` or `difference`.
            The default is None.
        main : bool, optional
            If True, the values of the benchmarked QAOA object will be plotted.
            The default is True.
        reference : bool, optional
            If True, the values of the reference QAOA object will be plotted.
            The default is False.
        difference : bool, optional
            If True, the difference between the values of the benchmarked QAOA object and the values of the reference QAOA object will be plotted.
            The default is False.
        one_plot : bool, optional
            If True, the values of the benchmarked QAOA object,
            the values of the reference QAOA object and the difference between the values of the benchmarked QAOA object
            and the values of the reference QAOA object will be plotted in the same plot.
            It is not available if the sweep is over more than one parameter.
            The default is False.
        plot_options : dict, optional
            The options for the plot (plt.plot for one parameter, plt.pcolorfast for two parameters).
        """

        # check the inputs
        assert (
            isinstance(ax, plt.Axes) or ax is None
        ), "ax must be a matplotlib Axes or None"
        assert (
            isinstance(labels, list) or labels is None
        ), "labels must be a list of two strings or None"
        assert (
            all([isinstance(label, str) for label in labels])
            if labels is not None
            else True
        ), "labels must be a list of two strings or None"
        assert (
            len(labels) == 2 if labels is not None else True
        ), "labels must be a list of two strings, one for each axis"
        for plot in ["main", "reference", "difference"]:
            assert isinstance(eval(plot), bool), plot + " must be a boolean"
        assert (
            main or reference or difference
        ), "You must specify at least one of the main, reference or difference plots"
        assert isinstance(plot_options, dict), "plot_options must be a dictionary"

        # if labels_legend is not a list or tuple, make it a list
        if not isinstance(labels_legend, (list, tuple)):
            labels_legend = [labels_legend for _ in range(3)]

        # create a dictionary where the keys are the three possible plots and the values are a tuple
        # with the boolean (which says if the plot is requested) and the values to plot
        values = {
            "main": (main, self.values if main else None),
            "reference": (reference, self.values_reference if reference else None),
            "difference": (difference, self.difference if difference else None),
        }

        # create the figure and axes if the axis are not provided
        ax_input = ax  # save the input axis
        nrows = 1
        if ax_input is None:
            nrows = (
                1 if one_plot else sum([1 if values[key][0] else 0 for key in values])
            )
            fig, ax = plt.subplots(nrows=nrows, ncols=1, figsize=(6.5, 5 * nrows))

        # plot the requested plots
        count_sp = 0  # counter of subplots
        for key in values:
            if values[key][0]:  # skip the plot if it is not requested
                # raise an exception if the values for a plo requested are not available
                if values[key][1] is None:
                    raise Exception(
                        "You must run the benchmark before plotting the results, there are no values for the "
                        + key
                        + " plot"
                    )

                # get the ranges to use, use the reference ranges if the key is "reference", otherwise use the main ranges
                ranges = self.ranges if key != "reference" else self.ranges_reference
                ranges_to_use = [r for r in ranges if len(r) == 2]
                n_params = len(ranges_to_use)

                # create the axes, the range is provided by the ranges_to_use and the number of values is provided by the values[key][1].shape
                axes = [
                    np.linspace(r[0], r[1], values[key][1].shape[i])
                    for i, r in enumerate(ranges_to_use)
                ]

                # skip the plot if there are more than two parameters to plot, and print a warning
                if n_params > 2:
                    print(
                        "Only 1 or 2 parameters can be plotted using this method. You are trying to plot "
                        + str(n_params)
                        + " parameters. You can use the argument params_to_plot to specify which parameters to plot."
                    )
                    break
                # if verbose, print the parameters used for the plot
                elif verbose:
                    print(
                        "Plotting the " + key + " plot with the following parameters:"
                    )
                    for i, r in enumerate(ranges):
                        if len(r) == 2:
                            print(
                                "\tParameter "
                                + str(i)
                                + ": "
                                + str(r[0])
                                + " to "
                                + str(r[1])
                                + ", with "
                                + str(values[key][1].shape[i])
                                + " values"
                            )
                        else:
                            print("\tParameter " + str(i) + ": " + str(r[0]))

                ## plot the values

                # if we are plotting only one plot, use the same axes, otherwise different axes for each subplot
                axis = ax if nrows == 1 else ax[count_sp]

                # set the labels and plot the values
                if n_params == 1:
                    axis.set_xlabel(
                        "Parameter {}".format(
                            [i for i, r in enumerate(ranges) if len(r) == 2][0]
                        )
                    )
                    axis.set_ylabel("Expectation value")

                    label_legend = (
                        labels_legend[count_sp]
                        if labels_legend[count_sp] is not None
                        else key
                    )
                    axis.plot(*axes, values[key][1], label=label_legend, **plot_options)
                else:
                    if one_plot:
                        raise Exception(
                            "For two parameters, you must specify one_plot=False"
                        )

                    axis.set_xlabel(
                        "Parameter {}".format(
                            [i for i, r in enumerate(ranges) if len(r) == 2][0]
                        )
                    )
                    axis.set_ylabel(
                        "Parameter {}".format(
                            [i for i, r in enumerate(ranges) if len(r) == 2][1]
                        )
                    )
                    plot = axis.pcolorfast(*axes, values[key][1], **plot_options)
                    _ = plt.colorbar(plot)

                # set the title
                if not one_plot and title is None:
                    axis.set_title(key + " plot")
                elif not title is None:
                    if not isinstance(title, list) and not isinstance(title, tuple):
                        axis.set_title(title)
                    else:
                        if len(title) < nrows:
                            raise Exception(
                                "If title is a list or tuple, it must have the same length as the number of plots"
                            )
                        axis.set_title(title[count_sp])

                # replace the labels if specified
                if labels is not None:
                    axis.set_xlabel(labels[0])
                    axis.set_ylabel(labels[1])

                # increase the counter for the axes
                count_sp += 1

        # add legend if one_plot
        if one_plot:
            ax.legend()

        # show the figure if the axis were not provided
        if ax_input is None:
            if nrows > 1:
                fig.subplots_adjust(hspace=0.3)  # add some space between the subplots
            plt.show()
            plt.close(fig)

    @staticmethod
    def __ordered_points(n_params, n_points_axis):
        """
        This method creates a grid of points for each parameter, and then creates all the combinations of these points.
        The points are ordered such that we evaluate the grid by rounds, the first round the resolution is very low,
        next rounds the resolution is higher,
        and also, every n_points = 2**n_params one point is used at each round, and the points are shuffled.

        It works very well for n_points_axis = 2**k + 1 (where k is any integer), but it also works for other values.

        Parameters
        ----------
        n_params : int
            The number of parameters to sweep.
        n_points_axis : int
            The number of points per axis.

        Returns
        -------
        points : list
            The list of points to evaluate in order.
        """

        assert (
            isinstance(n_params, int) and n_params > 0
        ), "The number of parameters must be an integer, and greater than 0"
        assert (
            isinstance(n_points_axis, int) and n_points_axis > 3
        ), "The number of points per axis must be an integer, and greater than 3"

        ## we create a grid of points for each axis
        axis_points = list(range(n_points_axis))

        ## first we create all the points to evaluate and we order them such that every n_points = 2**n_params one point is used at each round, and the points are shuffled

        # we separate the points in two lists, [0, 2, 4, ...] and [1, 3, 5, ...]
        axis_points_separated = (
            [[i] for i in axis_points[::2]],
            [[i] for i in axis_points[1::2]],
        )

        # we create a list of lists, where each list will tell how to combine the two lists of points to create the points for each round
        zero_one = [[0], [1]]
        order_list = zero_one
        for _ in range(n_params - 1):
            order_list = [order + y for order in order_list for y in zero_one]

        # the variable points will be a list of lists, where each list is a round of points to evaluate
        points = [
            axis_points_separated[order[0]] for order in order_list
        ]  # we start with the first axis

        # we create the points for each round by combining the points of the previous round with the points of the next axis
        for k in range(1, n_params):
            for x in range(len(points)):
                points[x] = [
                    point_i + point_j
                    for point_i in points[x]
                    for point_j in axis_points_separated[order_list[x][k]]
                ]

        # we shuffle the points at each round
        for i in range(len(points)):
            shuffle(points[i])

        # the final list of points to evaluate is a concatenation of all the rounds
        ordered_points = []
        for round in range(len(points)):
            ordered_points += points[round]

        ## then we reorder the points such that the first round we have a grid of points with very low resolution, and then at each round we increase the resolution
        values = np.zeros([n_points_axis for _ in range(n_params)])

        used_axis_points = []
        reordered_points = []
        for round in range(1, int(log2(len(axis_points) - 1)) + 1):
            new = [
                k
                for k in axis_points[:: (len(axis_points) - 1) // 2**round]
                if k not in used_axis_points
            ]
            used_axis_points += new

            points_ = ordered_points.copy()
            for point in ordered_points:
                if (
                    all([i in used_axis_points for i in point])
                    and values[tuple(point)] == 0
                ):
                    values[tuple(point)] = round
                    reordered_points.append(points_.pop(points_.index(point)))

            ordered_points = points_  # there are less points now

        return reordered_points
